@@ -1,6 +1,7 @@
+use crate::leapiter::{Iter, IterMut, OwnedIter};
 use crate::leapref::{Ref, RefMut};
 use crate::util::{allocate, deallocate, round_to_pow2, AllocationKind};
-use crate::{make_hash, MurmurHasher, Value};
+use crate::{make_hash, Value};
 use atomic::Atomic;
 use std::alloc::{Allocator, Global};
 use std::borrow::Borrow;
@@ -9,68 +10,61 @@ use std::sync::atomic::{
     AtomicBool, AtomicPtr, AtomicU32, AtomicU64, AtomicU8, AtomicUsize, Ordering,
 };
 
-/// The type used for hashed keys.
-type HashedKey = u64;
-
-/// The atomic type of the hashed key.
-type AtomicHashedKey = AtomicU64;
-
-/// Result types when an insert into a [LeapMap].
-pub(super) enum ConcurrentInsertResult<V> {
-    /// The insertion found a cell for the key, replaced the old value with
-    /// a new one, and returned the old value.
-    Replaced(V),
-    /// The insertion was performed with a new key, so a new cell was filled.
-    NewInsert,
-    /// The insertion found a cell whose value was stale, so the map data needs
-    /// to be migrated, after which the value should be inserted into the migrated
-    /// map.
-    Migration(V),
-    /// No new cell was found for the key withing the linear search range,
-    /// so we overflowed the max delta value and need to move the map to
-    /// a larger table.
-    Overflow(usize),
-}
+/// The default hasher for a [`LeapMap`].
+pub(crate) type DefaultHash = std::collections::hash_map::DefaultHasher;
 
 /// A concurrent hash map implementation which uses a modified form of RobinHood/
-/// Hopscotch probing. This implementation is lock-free, and therefore it
-/// *not possible* to deadlock.
+/// Hopscotch probing. This implementation is lock-free, and therefore it will
+/// *not* deadlock when any of the map operations are performed concurrently.
+/// The map is lock-free if the key and value types have built-in atomic support,
+/// and if not, an efficient spin-lock is used.
 ///
-/// By default, the `LeapMap` map uses Murmur for hahsing, see [MurmurHasher],
-/// however, it's simple enough to change the hasher, just create the map
-/// using `with_hasher` or `with_capacity_and_hasher`.
+/// By default, the `LeapMap` map uses the default hasher from the standard
+/// library, which is DOS resistent, but is less efficient. Any other hasher
+/// can be used instead, likely with improved performance but less security.
+/// Using a different hasher is as simple a reating the map using the
+/// `with_hasher` or `with_capacity_and_hasher` methods.
 ///
-/// This requires that the value type implement [Value], which is very simple
-/// to do, and requires that a two values of the value type to be stored to be
-/// used as a null and redirect flags. It's best to choose something that is
-/// simple and efficeint to evaluate. For example, u64 uses u64::MAX and u64::MAx -1,
-/// respectively.
+/// Values stored in the map need to mplement the [`Value`] trait. The trait is
+/// simple to implement, requiring that two values be reserved, one for the
+/// `null` case, which is used as a placeholder, and one for a `redirect`, which
+/// is used to signal that the map is being migrated. This trait is only implemented
+/// for numeric types and uses MAX and MAX -1 for the null and redirect values,
+/// respectively. For any other type this needs to be implemented.
+///
+/// This map is very fast for concurrent operations since the locking is either
+/// atomic or uses very finer grained locking than other maps.
 ///
 /// # Limitations
 ///
-/// This hash map *does not* store keys, so if that is required, then another
-/// map would be better.
+/// This biggest limitations of this map are that the interface is slightly
+/// different than using [`std::sync::RwLock<HashMap>`]. The type returned
+/// when calling `get`, `get_mut`, `iter`, etc, are not references, but rather a
+/// [`Ref`] or [`RefMut`] type which acts like a reference but still allows concurrent
+/// operations on both that reference type and the map. This interface is still
+/// being worked out, and might be changed in the future.
 ///
-/// Currently, moving the table *does not* use a custom allocator, which would
-/// further improve the implementation.
+/// When adding a key-value pair to the map, it's possible that the map is too
+/// full and there is no bucket, in which case a larger map needs to be allocated
+/// and then filled from the old map. This will happen concurrently. The current
+/// implementation *does not* use a custom allocator, but would likely further
+/// improve the performance.
 ///
-/// The number of elements in the map must be a power of two.
+/// The number of elements in the map must be a power of two. This is handled
+/// internally.
 ///
-/// The biggest difference between this map and the std HashMap with a RwLock,
-/// or something similatr, is the type returned when calling `get` and `get_mut`.
-/// Here, we return a reference type to the value in the map, however, the type
-/// is atomic, and therefore needs to be modified and accessed atomically. While
-/// it would be possible to give a reference to the value, it would require
-/// locking the map, which is inefficient for concurrent use. We therefore
-/// return [Ref] or [RefMut] types, which provide methods to safely and efficiently
-/// modify the value. Altrenatively, a better way to modify values in the map is
-/// to simply use `update`.
+/// Lastly, this map stores additional data for each key-value pair. There are
+/// 8 bytes used to store offsets for the probe search, and 8 bytes for the hashed
+/// key. This is quite a lot of overhead, but this map is designed for good performace
+/// and scaling. This will be a focus area for improvement in the future, but if
+/// memory use is more important that performance, another map will be a better
+/// choice.
 ///
 /// # Threading
 ///
-/// The map *is* thread-safe, and items cana be added, removed, and updated from
-/// any number of threads concurrently.
-pub struct LeapMap<K, V, H = BuildHasherDefault<MurmurHasher>, A: Allocator = Global> {
+/// The map *is* thread-safe, and items can be added, removed, pdated, and iterated
+/// over cuncurrently from any number of threads.
+pub struct LeapMap<K, V, H = BuildHasherDefault<DefaultHash>, A: Allocator = Global> {
     /// Pointer to the buckets for the map. This needs to be a pointer to that
     /// we can atomically get the buckets for operations, since when the map
     /// is resized the buckets might change, and we need to ensure that all
@@ -101,16 +95,21 @@ impl<'a, K, V, H, A: Allocator> LeapMap<K, V, H, A> {
     fn get_table_mut(&self, ordering: Ordering) -> &'a mut Table<K, V> {
         unsafe { &mut *self.table.load(ordering) }
     }
+
+    /// Returns true if the table has been allocated.
+    fn is_allocated(&self) -> bool {
+        !self.table.load(Ordering::Relaxed).is_null()
+    }
 }
 
-impl<K, V> LeapMap<K, V, BuildHasherDefault<MurmurHasher>, Global>
+impl<K, V> LeapMap<K, V, BuildHasherDefault<DefaultHash>, Global>
 where
-    K: Eq + Hash + Clone,
+    K: Eq + Hash + Copy,
     V: Value,
 {
     /// Creates the hash map with space for the default number of elements, which
     /// will use the global allocator for allocation of the map data.
-    pub fn new() -> LeapMap<K, V, BuildHasherDefault<MurmurHasher>, Global> {
+    pub fn new() -> LeapMap<K, V, BuildHasherDefault<DefaultHash>, Global> {
         Self::new_in(Global)
     }
 
@@ -118,10 +117,10 @@ where
     /// the global allocator for allocation of the map data.
     pub fn with_capacity(
         capacity: usize,
-    ) -> LeapMap<K, V, BuildHasherDefault<MurmurHasher>, Global> {
+    ) -> LeapMap<K, V, BuildHasherDefault<DefaultHash>, Global> {
         Self::with_capacity_and_hasher_in(
             capacity,
-            BuildHasherDefault::<MurmurHasher>::default(),
+            BuildHasherDefault::<DefaultHash>::default(),
             Global,
         )
     }
@@ -129,7 +128,7 @@ where
 
 impl<K, V, H> LeapMap<K, V, H, Global>
 where
-    K: Eq + Hash + Clone,
+    K: Eq + Hash + Copy,
     V: Value,
     H: BuildHasher + Default,
 {
@@ -143,7 +142,7 @@ where
 
 impl<'a, K, V, H, A> LeapMap<K, V, H, A>
 where
-    K: Eq + Hash + Clone + 'a,
+    K: Eq + Hash + Copy + 'a,
     V: Value + 'a,
     H: BuildHasher + Default,
     A: Allocator,
@@ -180,6 +179,7 @@ where
         let migrator = unsafe { &mut *migrator_ptr };
         migrator.initialize();
 
+        let capacity = round_to_pow2(capacity.max(Self::INITIAL_SIZE));
         let table_ptr = Self::allocate_and_init_table(&allocator, capacity);
         LeapMap {
             table: AtomicPtr::new(table_ptr),
@@ -206,7 +206,7 @@ where
     /// let map = leapfrog::LeapMap::new();
     /// map.insert(0, 12);
     /// if let Some(mut r) = map.get(&0) {
-    ///     assert_eq!(r.value(), 12);
+    ///     assert_eq!(r.value(), Some(12));
     /// } else {
     ///     // Map is stale
     ///     assert!(false);
@@ -219,8 +219,7 @@ where
         Q: Hash + Eq,
     {
         let hash = make_hash::<K, Q, H>(&self.hash_builder, key);
-        self.find(hash)
-            .map_or(None, |cell| Some(Ref::new(self, cell, hash)))
+        self.find(key, hash).map(|cell| Ref::new(self, cell, hash))
     }
 
     /// Returns a mutable reference type to the value corresponding to the `key`.
@@ -237,7 +236,7 @@ where
     ///        // Map is being moved, can't update
     ///     }
     /// }
-    /// assert_eq!(map.get(&1).unwrap().value(), 14);
+    /// assert_eq!(map.get(&1).unwrap().value(), Some(14));
     /// assert!(map.get(&2).is_none());
     ///```
     pub fn get_mut<Q: ?Sized>(&'a self, key: &Q) -> Option<RefMut<'a, K, V, H, A>>
@@ -246,8 +245,8 @@ where
         Q: Hash + Eq,
     {
         let hash = make_hash::<K, Q, H>(&self.hash_builder, key);
-        self.find(hash)
-            .map_or(None, |cell| Some(RefMut::new(self, cell, hash)))
+        self.find(key, hash)
+            .map(|cell| RefMut::new(self, cell, hash))
     }
 
     /// Returns true if the map contains the specified `key`.
@@ -285,7 +284,7 @@ where
         Q: Hash + Eq,
     {
         let hash = make_hash::<K, Q, H>(&self.hash_builder, key);
-        self.find(hash).map_or(None, |cell| self.erase_value(cell))
+        self.find(key, hash).and_then(|cell| self.erase_value(cell))
     }
 
     /// Updates a key-value pair.
@@ -299,15 +298,19 @@ where
     ///
     /// ```
     /// let map = leapfrog::LeapMap::new();
-    /// assert_eq!(map.insert(37, 12), None);
-    /// assert_eq!(map.update(37, 14), Some(12));
+    /// assert_eq!(map.insert(&37, 12), None);
+    /// assert_eq!(map.update(&37, 14), Some(12));
     /// ```
-    pub fn update(&self, key: K, value: V) -> Option<V> {
-        let hash = make_hash::<K, _, H>(&self.hash_builder, &key);
+    pub fn update<Q: ?Sized>(&self, key: &Q, value: V) -> Option<V>
+    where
+        K: Borrow<Q>,
+        Q: Hash + Eq,
+    {
+        let hash = make_hash::<K, _, H>(&self.hash_builder, key);
         debug_assert!(!value.is_null());
 
         loop {
-            match self.find(hash) {
+            match self.find(key, hash) {
                 Some(cell) => {
                     let old = cell.value.load(Ordering::Relaxed);
                     match Self::exchange_value(cell, value, old) {
@@ -320,7 +323,7 @@ where
                         }
                         ConcurrentInsertResult::Overflow(_) => {
                             // This should never happen, so we panic if it does.
-                            assert!(false);
+                            panic!("LeapMap update overflowed");
                         }
                         ConcurrentInsertResult::Migration(_) => {
                             // Help with the migration and then try again
@@ -363,23 +366,20 @@ where
             let size_mask = table.size_mask;
             let buckets = table.bucket_slice_mut();
 
-            match Self::insert_or_find(hash, value, buckets, size_mask) {
+            match Self::insert_or_find(hash, &key, value, buckets, size_mask) {
                 ConcurrentInsertResult::Overflow(overflow_index) => {
                     // Try and create the table migration. Only a single thread
                     // should do this. Any threads which don't get the flag should
                     // start cleaning up old tables which have not been deallocated.
-                    let mut migrator = unsafe { &mut *self.migrator.load(Ordering::Relaxed) };
+                    let migrator = unsafe { &mut *self.migrator.load(Ordering::Relaxed) };
                     if migrator.set_initialization_begin_flag() {
                         Self::initialize_migrator(
                             &self.allocator,
-                            &mut migrator,
+                            migrator,
                             &self.table,
                             overflow_index,
                         );
                         migrator.set_initialization_complete_flag();
-                    } else {
-                        // Try and clean up any old migration tables.
-                        //migrator.cleanup_stale_table(&self.allocator);
                     }
 
                     // Migration is created, all threads can see that try and migrate.
@@ -403,6 +403,66 @@ where
         }
     }
 
+    /// Creates an iterator over a [`LeapMap`] which yields immutable key-value
+    /// pairs.
+    ///
+    /// # Threading
+    ///
+    /// This is thread-safe and can be called from any number of threads, and
+    /// will *not* deadlock. See the [`Ref`] type for how to access iterated
+    /// values.
+    //
+    /// # Examples
+    ///
+    /// ```
+    /// use leapfrog::LeapMap;
+    ///
+    /// let map = LeapMap::new();
+    /// map.insert(12, 27);
+    /// assert_eq!(map.iter().count(), 1);
+    ///
+    /// for mut item in map.iter() {
+    ///     assert_eq!(item.key_value(), Some((12, 27)));
+    /// }
+    /// ```
+    pub fn iter(&'a self) -> Iter<'a, K, V, H, A> {
+        Iter::new(self)
+    }
+
+    /// Creates an iterator over a [`LeapMap`] which yields mutable key-value
+    /// pairs.
+    ///
+    /// # Threading
+    ///
+    /// This is thread-safe and can be called from any number of threads, and
+    /// will *not* deadlock. See the [`Ref`] type for how to access iterated
+    /// values.
+    //
+    /// # Examples
+    ///
+    /// ```
+    /// use leapfrog::LeapMap;
+    ///
+    /// let map = LeapMap::new();
+    /// map.insert(12u32, 27u32);
+    /// map.insert(13u32, 28u32);
+    /// assert_eq!(map.iter_mut().count(), 2);
+    ///
+    /// for mut item in map.iter_mut() {
+    ///     let old = item.update(|v| {
+    ///         let current = *v;
+    ///         *v = current + current;
+    ///     });
+    ///     assert!(old.is_some());
+    /// }
+    ///
+    /// assert_eq!(map.get(&12).unwrap().value(), Some(54));
+    /// assert_eq!(map.get(&13).unwrap().value(), Some(56));
+    /// ```
+    pub fn iter_mut(&'a self) -> IterMut<'a, K, V, H, A> {
+        IterMut::new(self)
+    }
+
     /// Returns the length of the map.
     ///
     /// This computes the length each time, since the alternative--pdating the
@@ -435,17 +495,26 @@ where
         elements
     }
 
+    /// Returns `true` if the map contains no elements.
+    pub fn is_empty(&self) -> bool {
+        self.len() == 0
+    }
+
     /// Tries to find the value for the `hash`, without inserting into the map.
     /// This will reuturn a reference to the cell if the find succeeded.
-    pub(crate) fn find(&self, hash: HashedKey) -> Option<&'a AtomicCell<V>> {
-        debug_assert!(hash != Self::null_hash());
+    pub(crate) fn find<Q: ?Sized>(&self, key: &Q, hash: HashedKey) -> Option<&'a AtomicCell<K, V>>
+    where
+        K: Borrow<Q>,
+        Q: Hash + Eq,
+    {
+        debug_assert!(hash != null_hash());
 
         loop {
             let table = self.get_table(Ordering::Acquire);
             let size_mask = table.size_mask;
             let buckets = table.bucket_slice();
 
-            if let Some(cell) = self.find_inner(hash, buckets, size_mask) {
+            if let Some(cell) = self.find_inner(key, hash, buckets, size_mask) {
                 let value = cell.value.load(Ordering::Acquire);
                 if !value.is_redirect() {
                     if value.is_null() {
@@ -467,18 +536,24 @@ where
     /// `buckets`, which have an assosciated `size_mask` representing the number
     /// of cells in the buckets. This returns a reference to the cell if the
     /// find was successful.
-    fn find_inner(
+    fn find_inner<Q: ?Sized>(
         &self,
+        key: &Q,
         hash: HashedKey,
         buckets: &'a [Bucket<K, V>],
         size_mask: usize,
-    ) -> Option<&'a AtomicCell<V>> {
+    ) -> Option<&'a AtomicCell<K, V>>
+    where
+        K: Borrow<Q>,
+        Q: Hash + Eq,
+    {
         let mut index = hash as usize & size_mask;
         let cell = get_cell(buckets, index, size_mask);
         let probe_hash = cell.hash.load(Ordering::Relaxed);
-        if hash == probe_hash {
-            return Some(&cell);
-        } else if probe_hash == Self::null_hash() {
+        let probe_key = cell.key.load(Ordering::Relaxed);
+        if hash == probe_hash && key.eq(probe_key.borrow()) {
+            return Some(cell);
+        } else if probe_hash == null_hash() {
             return None;
         }
 
@@ -488,12 +563,13 @@ where
             index = (index + delta as usize) & size_mask;
             let cell = get_cell(buckets, index, size_mask);
             let probe_hash = cell.hash.load(Ordering::Relaxed);
+            let probe_key = cell.key.load(Ordering::Relaxed);
 
             // NOTE: It's possible that a concurrent insert memory reordering
             //       causes probe_hash to be default, but we don't need to check
             //       for that case, and we can just follow the probe chain.
-            if probe_hash == hash {
-                return Some(&cell);
+            if probe_hash == hash && key.eq(probe_key.borrow()) {
+                return Some(cell);
             }
 
             delta = get_second_delta(buckets, index, size_mask).load(Ordering::Relaxed);
@@ -511,7 +587,7 @@ where
     /// value, or modify any deltas for the probe chain, since that would break
     /// lookups of other cells in the map. This also means that if another cell
     /// is inserted with the same hash then there is a faster path when inserting.
-    fn erase_value(&self, mut cell: &'a AtomicCell<V>) -> Option<V> {
+    fn erase_value(&self, mut cell: &'a AtomicCell<K, V>) -> Option<V> {
         loop {
             let value = cell.value.load(Ordering::Relaxed);
 
@@ -522,6 +598,7 @@ where
                 return None;
             }
 
+            let key = cell.key.load(Ordering::Relaxed);
             match cell.value.compare_exchange(
                 value,
                 V::null(),
@@ -545,7 +622,7 @@ where
                     let hash = cell.hash.load(Ordering::Relaxed);
                     loop {
                         self.participate_in_migration();
-                        match self.find(hash) {
+                        match self.find(&key, hash) {
                             Some(new_cell) => {
                                 cell = new_cell;
                                 // This should almost always be true, migration is finished.
@@ -600,7 +677,7 @@ where
         let new_table_size = estimated.max(Self::INITIAL_SIZE as usize);
 
         // Migrator initialization:
-        let dst_table_ptr = Self::allocate_and_init_table(&allocator, new_table_size);
+        let dst_table_ptr = Self::allocate_and_init_table(allocator, new_table_size);
         let dst_table = unsafe { &*dst_table_ptr };
         debug_assert!(dst_table.size_mask != 0);
         migrator.dst_table.store(dst_table_ptr, Ordering::Relaxed);
@@ -626,9 +703,10 @@ where
             let index = migrator.stale_source_index((last_source + i) as usize);
 
             // If this is not null, then we have wrapped:
-            debug_assert!(
-                migrator.stale_sources[index].load(Ordering::Relaxed) == std::ptr::null_mut()
-            );
+            // FIXME: This occasionally causes a panic in the tests!
+            debug_assert!(migrator.stale_sources[index]
+                .load(Ordering::Relaxed)
+                .is_null());
 
             migrator.stale_sources[index]
                 .store(source.table.load(Ordering::Relaxed), Ordering::Relaxed);
@@ -639,7 +717,7 @@ where
             table: AtomicPtr::new(src_table_ptr),
             index: AtomicUsize::new(0),
         };
-        if migrator.sources.len() < 1 {
+        if migrator.sources.is_empty() {
             migrator.sources.push(source);
         } else {
             migrator.sources[0] = source;
@@ -802,7 +880,7 @@ where
 
             // Create the new table for the migrator:
             let new_dst_table_ptr =
-                Self::allocate_and_init_table(&allocator, (table.size_mask + 1) * 2);
+                Self::allocate_and_init_table(allocator, (table.size_mask + 1) * 2);
             migrator
                 .dst_table
                 .store(new_dst_table_ptr, Ordering::Relaxed);
@@ -852,6 +930,7 @@ where
     /// a migration is in progress.
     pub(super) fn insert_or_find(
         hash: HashedKey,
+        key: &K,
         value: V,
         buckets: &[Bucket<K, V>],
         size_mask: usize,
@@ -863,14 +942,17 @@ where
         let cell = get_cell(buckets, index, size_mask);
         let mut probe_hash = cell.hash.load(Ordering::Relaxed);
 
-        if probe_hash == Self::null_hash() {
+        if probe_hash == null_hash() {
             match cell
                 .hash
                 .compare_exchange(probe_hash, hash, Ordering::Relaxed, Ordering::Relaxed)
             {
                 Ok(_) => {
-                    // Succeeded in setting the hash and claiming the cell, try
-                    // and exchange the value:
+                    // This is an empty cell, and we won the race to claim it.
+                    // We can just set the key. The only problem we could have
+                    // here is a migration, in which case the value exchange below
+                    // will fail.
+                    cell.key.store(*key, Ordering::Relaxed);
                     return Self::exchange_value(cell, value, V::null());
                 }
                 Err(new_hash) => {
@@ -880,8 +962,10 @@ where
             }
         }
 
-        // Hash already in table, exchange and return the value.
-        if probe_hash == hash {
+        // Hash already in table, check if the keys match, if they do then we
+        // can exchange and return the value.
+        let probe_key = cell.key.load(Ordering::Relaxed);
+        if probe_hash == hash && probe_key.eq(key.borrow()) {
             let old_value = cell.value.load(Ordering::Acquire);
 
             // Check if the table is being moved:
@@ -893,6 +977,7 @@ where
             return Self::exchange_value(cell, value, old_value);
         }
 
+        // Hard part, need to find a new cell ...
         // Follow the linked probes to find a new cell.
         // Here, we get the offset delta which will allow us to find the desired
         // bucket. After that, we need to use the second set of delta values to
@@ -914,23 +999,24 @@ where
 
                 let cell = get_cell(buckets, index, size_mask);
                 let mut probe_hash = cell.hash.load(Ordering::Relaxed);
-                if probe_hash == Self::null_hash() {
+                if probe_hash == null_hash() {
                     // The cell has been linked to the probe chain, but the hash
                     // is not yet visible. We chould change this here to use
                     // acquire and release, which would guarantee that the change
-                    // is visible, but it's easier to just poll: FIXME
+                    // is visible, but it's easier to just poll:
+                    // FIXME : Cange this to acquire releaase ...
                     loop {
                         probe_hash = cell.hash.load(Ordering::Acquire);
-                        if probe_hash != Self::null_hash() {
+                        if probe_hash != null_hash() {
                             break;
                         }
                     }
                 }
 
-                // Only hashes in the same bucket can be linked:
+                // Only hashes in the same bucket can be linked.
                 debug_assert!((probe_hash ^ hash) as usize & size_mask == 0);
-
-                if probe_hash == hash {
+                let probe_key = cell.key.load(Ordering::Relaxed);
+                if probe_hash == hash && probe_key.eq(key.borrow()) {
                     let old_value = cell.value.load(Ordering::Acquire);
 
                     // Check if the table is being moved:
@@ -942,7 +1028,7 @@ where
                 }
 
                 // We updated the index and the delta value, so we will go back
-                // to the start of the loop and search the next probe in the chain ...
+                // to the start of the loop and search the next probe in the chain
             } else {
                 // We have reached the end of the probe chain for the current
                 // bucket. We now need to move to linear probing until we either
@@ -959,7 +1045,7 @@ where
                     // to the previous cell in the same bucket.
                     let cell = get_cell(buckets, index, size_mask);
                     let mut probe_hash = cell.hash.load(Ordering::Relaxed);
-                    if probe_hash == Self::null_hash() {
+                    if probe_hash == null_hash() {
                         // Same as above, there is an empty cell, we we want to
                         // try and reserve it. If we lose the race here, then
                         // we just fall through to the next stage ...
@@ -973,6 +1059,10 @@ where
                                 // Now link the cell to the previous cell in the same bucket:
                                 let offset = (index - prev_link_index) as u8;
                                 prev_link.store(offset, Ordering::Relaxed);
+
+                                // CHECK: No race here with another cell
+                                cell.key.store(*key, Ordering::Relaxed);
+
                                 return Self::exchange_value(cell, value, V::null());
                             }
                             Err(updated) => probe_hash = updated,
@@ -981,7 +1071,8 @@ where
 
                     // Check for the same hash, so we can replace:
                     let x = hash ^ probe_hash;
-                    if x == 0 {
+                    let probe_key = cell.key.load(Ordering::Relaxed);
+                    if x == 0 && probe_key == *key {
                         let old_value = cell.value.load(Ordering::Acquire);
                         return Self::exchange_value(cell, value, old_value);
                     }
@@ -1031,7 +1122,11 @@ where
     /// If the exchange returns a redirect value then this will return a
     /// [ConcurrentInsertResult::Migration] with the `desired` value, which
     /// needs to be inserted into the new table.
-    fn exchange_value(cell: &AtomicCell<V>, desired: V, old_value: V) -> ConcurrentInsertResult<V> {
+    fn exchange_value(
+        cell: &AtomicCell<K, V>,
+        desired: V,
+        old_value: V,
+    ) -> ConcurrentInsertResult<V> {
         let exchange_result =
             cell.value
                 .compare_exchange(old_value, desired, Ordering::AcqRel, Ordering::Relaxed);
@@ -1069,30 +1164,31 @@ where
         assert!(cells >= 4 && (cells % 2 == 0));
         let bucket_count = cells >> 2;
         let bucket_ptr =
-            allocate::<Bucket<K, V>, A>(&allocator, bucket_count, AllocationKind::Uninitialized);
+            allocate::<Bucket<K, V>, A>(allocator, bucket_count, AllocationKind::Uninitialized);
         let buckets = unsafe { std::slice::from_raw_parts_mut(bucket_ptr, bucket_count) };
 
         // Since AtomicU8 and AtomicU64 are the same as u8 and u64 in memory,
         // we can write them as zero, rather than calling the atomic versions
-        for i in 0..bucket_count {
+        for bucket in buckets.iter_mut().take(bucket_count) {
             unsafe {
-                let bucket_deltas = &mut buckets[i].deltas as *mut AtomicU8;
+                let bucket_deltas = &mut bucket.deltas as *mut AtomicU8;
                 std::ptr::write_bytes(bucket_deltas, 0, 8);
             };
 
             for cell in 0..4 {
                 unsafe {
-                    let cell_hash = &mut buckets[i].cells[cell].hash as *mut AtomicHashedKey;
+                    // We only need to write the hash as null, an can ignore th key.
+                    let cell_hash = &mut bucket.cells[cell].hash as *mut AtomicHashedKey;
                     std::ptr::write_bytes(cell_hash, 0, 1);
                 };
 
-                // FIXME: We should check if the stored type is directly writable ..
-                let cell_value = &mut buckets[i].cells[cell].value;
+                // FIXME: Check if the stored type is directly writable ..
+                let cell_value = &mut bucket.cells[cell].value;
                 *cell_value = Atomic::new(V::null());
             }
         }
 
-        let table_ptr = allocate::<Table<K, V>, A>(&allocator, 1, AllocationKind::Uninitialized);
+        let table_ptr = allocate::<Table<K, V>, A>(allocator, 1, AllocationKind::Uninitialized);
         let table = unsafe { &mut *table_ptr };
 
         table.buckets = bucket_ptr;
@@ -1101,10 +1197,30 @@ where
         table_ptr
     }
 
-    /// Returns the null hash value.
-    #[inline]
-    const fn null_hash() -> HashedKey {
-        0 as HashedKey
+    pub(crate) fn get_cell_at_index_mut(&'a self, index: usize) -> Option<RefMut<'a, K, V, H, A>> {
+        let table = self.get_table(Ordering::Acquire);
+        if index >= table.size() {
+            return None;
+        }
+
+        let buckets = table.bucket_slice();
+        let size_mask = table.size_mask;
+        let cell = get_cell(buckets, index, size_mask);
+        let cell_hash = cell.hash.load(Ordering::Relaxed);
+        Some(RefMut::new(self, cell, cell_hash))
+    }
+
+    pub(crate) fn get_cell_at_index(&'a self, index: usize) -> Option<Ref<'a, K, V, H, A>> {
+        let table = self.get_table(Ordering::Acquire);
+        if index >= table.size() {
+            return None;
+        }
+
+        let buckets = table.bucket_slice();
+        let size_mask = table.size_mask;
+        let cell = get_cell(buckets, index, size_mask);
+        let cell_hash = cell.hash.load(Ordering::Relaxed);
+        Some(Ref::new(self, cell, cell_hash))
     }
 
     /// Calculates the number of remaining migration units for the `size_mask`.
@@ -1116,28 +1232,49 @@ where
 
 impl<K, V, H, A: Allocator> Drop for LeapMap<K, V, H, A> {
     fn drop(&mut self) {
-        let table = self.get_table_mut(Ordering::SeqCst);
+        if self.is_allocated() {
+            let table = self.get_table_mut(Ordering::SeqCst);
 
-        let bucket_ptr = table.buckets;
-        let bucket_count = table.size() >> 2;
-        deallocate::<Bucket<K, V>, A>(&self.allocator, bucket_ptr, bucket_count);
+            let bucket_ptr = table.buckets;
+            let bucket_count = table.size() >> 2;
+            deallocate::<Bucket<K, V>, A>(&self.allocator, bucket_ptr, bucket_count);
 
-        let table_ptr = self.table.load(Ordering::Relaxed);
-        deallocate::<Table<K, V>, A>(&self.allocator, table_ptr, 1);
-
-        let migrator_ptr = self.migrator.load(Ordering::SeqCst);
-        let migrator = unsafe { &*migrator_ptr };
-        while migrator.stale_tables_remaining() > 0 {
-            migrator.cleanup_stale_table(&self.allocator);
+            let table_ptr = self.table.load(Ordering::Relaxed);
+            deallocate::<Table<K, V>, A>(&self.allocator, table_ptr, 1);
         }
 
-        // We don't need to deallocate the buckets for the migrator, since that
-        // has already been handled.
         let migrator_ptr = self.migrator.load(Ordering::SeqCst);
-        deallocate::<Migrator<K, V>, A>(&self.allocator, migrator_ptr, 1);
+        if !migrator_ptr.is_null() {
+            let migrator = unsafe { &*migrator_ptr };
+            while migrator.stale_tables_remaining() > 0 {
+                migrator.cleanup_stale_table(&self.allocator);
+            }
+            deallocate::<Migrator<K, V>, A>(&self.allocator, migrator_ptr, 1);
+        }
     }
 }
 
+impl<K, V, H, A> IntoIterator for LeapMap<K, V, H, A>
+where
+    K: Eq + Hash + Copy,
+    V: Value,
+    H: BuildHasher + Default,
+    A: Allocator,
+{
+    type Item = (K, V);
+    type IntoIter = OwnedIter<K, V, H, A>;
+
+    fn into_iter(self) -> Self::IntoIter {
+        OwnedIter::new(self)
+    }
+}
+
+/// Returns the null hash value.
+#[inline]
+pub(crate) const fn null_hash() -> u64 {
+    0u64
+}
+
 /// Gets a reference to a cell using the `index` to find the cell in the
 /// `buckets` slice.
 #[inline]
@@ -1145,7 +1282,7 @@ fn get_cell<K, V: Value>(
     buckets: &[Bucket<K, V>],
     index: usize,
     size_mask: usize,
-) -> &AtomicCell<V> {
+) -> &AtomicCell<K, V> {
     let bucket_index = get_bucket_index(index, size_mask);
     let cell_index = get_cell_index(index);
     &buckets[bucket_index].cells[cell_index]
@@ -1206,6 +1343,29 @@ const fn get_cell_index(index: usize) -> usize {
     index & 3
 }
 
+/// The type used for hashed keys.
+type HashedKey = u64;
+
+/// The atomic type of the hashed key.
+type AtomicHashedKey = AtomicU64;
+
+/// Result types when an insert into a [LeapMap].
+pub(super) enum ConcurrentInsertResult<V> {
+    /// The insertion found a cell for the key, replaced the old value with
+    /// a new one, and returned the old value.
+    Replaced(V),
+    /// The insertion was performed with a new key, so a new cell was filled.
+    NewInsert,
+    /// The insertion found a cell whose value was stale, so the map data needs
+    /// to be migrated, after which the value should be inserted into the migrated
+    /// map.
+    Migration(V),
+    /// No new cell was found for the key withing the linear search range,
+    /// so we overflowed the max delta value and need to move the map to
+    /// a larger table.
+    Overflow(usize),
+}
+
 /// Underlying table for a [LeapMap]. This stores a pointer to the buckets and
 /// the mask for the number of cells which are stored in the table.
 struct Table<K, V> {
@@ -1239,9 +1399,11 @@ impl<K, V> Table<K, V> {
 /// Here we need the value to be atomic. Since V is generic, it's possible that
 /// there is no built-in atomic support for it, in which case the `Atomic`
 /// type will fall back to using an efficient spinlock implementation.
-pub struct AtomicCell<V> {
+pub struct AtomicCell<K, V> {
     /// The hashed value of the cell.
     pub(crate) hash: AtomicHashedKey,
+    /// The key for the cell.
+    pub(crate) key: Atomic<K>,
     /// The value assosciated with the hash.
     pub(crate) value: Atomic<V>,
 }
@@ -1257,9 +1419,7 @@ pub(super) struct Bucket<K, V> {
     /// delta value means.
     deltas: [AtomicU8; 8],
     /// Cells for the bucket.
-    cells: [AtomicCell<V>; 4],
-    /// Placeholder for the key.
-    _key: std::marker::PhantomData<K>,
+    cells: [AtomicCell<K, V>; 4],
 }
 
 /// A struct for a migration source, which stores a pointer to the source buckets
@@ -1352,11 +1512,7 @@ impl<K, V> Migrator<K, V> {
         let old = self
             .status
             .fetch_or(Self::INITIALIZATION_START_FLAG, Ordering::Relaxed);
-        if old & Self::INITIALIZATION_START_FLAG == 0 {
-            true // We set the flag
-        } else {
-            false
-        }
+        old & Self::INITIALIZATION_START_FLAG == 0
     }
 
     /// Sets the flag which indicates that the initialization process is complete.
@@ -1366,11 +1522,7 @@ impl<K, V> Migrator<K, V> {
         let old = self
             .status
             .fetch_or(Self::INITIALIZATION_END_FLAG, Ordering::Relaxed);
-        if old & Self::INITIALIZATION_END_FLAG == 0 {
-            true // We set the flag
-        } else {
-            false
-        }
+        old & Self::INITIALIZATION_END_FLAG == 0
     }
 
     /// If the migrator has completed/is completing the migration.
@@ -1447,7 +1599,7 @@ impl<K, V> Migrator<K, V> {
         end_index: usize,
     ) -> bool
     where
-        K: Hash + Eq + Clone,
+        K: Hash + Eq + Copy,
         H: BuildHasher + Default,
         A: Allocator,
         V: Value,
@@ -1469,7 +1621,7 @@ impl<K, V> Migrator<K, V> {
 
                 // Check if we have an unused cell, in which case we need to put
                 // a redirect marker in its place.
-                if src_hash == LeapMap::<K, V, H, A>::null_hash() {
+                if src_hash == null_hash() {
                     match cell.value.compare_exchange(
                         V::null(),
                         V::redirect(),
@@ -1513,6 +1665,8 @@ impl<K, V> Migrator<K, V> {
                         break;
                     }
 
+                    let src_key = cell.key.load(Ordering::Relaxed);
+
                     // We need to perform a migration of the cell. This should
                     // change. We actually want to get a reference to the cell
                     // here, and then try to exchange it here, rather than
@@ -1520,6 +1674,7 @@ impl<K, V> Migrator<K, V> {
                     loop {
                         match LeapMap::<K, V, H, A>::insert_or_find(
                             src_hash,
+                            &src_key,
                             src_value,
                             dst_buckets,
                             dst_size_mask,
@@ -1559,10 +1714,13 @@ impl<K, V> Migrator<K, V> {
                             }
                         }
                     }
+
+                    // Migrated the cell, proceed to next cell:
+                    break;
                 }
 
                 // Migrated the cell, proceed to next cell:
-                break;
+                //break;
             }
         }
 
